/*
 * MultiRAEPlus.java
 *
 * Created on October 27, 2005, 11:16 AM
 */

package sensor.raeSystems.multiRAEPlus;

//import com.rovlab.core.ModuleBase;
import sensor.PeriodicEventSource;
import sensor.PeriodicEventSourceImpl;
import sensor.raeSystems.InstantGasLevelsListener;
import util.Watchdog;
import util.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import util.StringSplit;

public class MultiRAEPlus extends PeriodicEventSourceImpl { //extends ModuleBase {
    public static final long TIMEOUT = 750; //ms
    
    public static final int CALIBRATION_ERROR = 1;
    public static final int ALARM_ERROR = 2;
    public static final int FAIL_ERROR = 4;
    public static final int HIGH_ERROR = 8;
    public static final int LOW_ERROR = 16;
    public static final int TWA_ERROR = 32;
    public static final int STEL_ERROR = 64;
    public static final int DRIFT_ERROR = 128;
    
    InputStream is;
    OutputStream os;
    BufferedReader r;
    Watchdog watchdog = new Watchdog();
    String[] sensorNames;
    
    /** Creates a new instance of MultiRAEPlus */
    public MultiRAEPlus(InputStream is, OutputStream os) throws IOException {
        this.is = is;
        this.os = os;
        try {
        r = new BufferedReader(new InputStreamReader(is));
        sensorNames = getSensorNamesFromUnit();
        } catch (IOException e) {
            stop(); //Close the streams
            throw (e);
        }
    }
    
    /**
     * Getter for property firmwareVersion.
     * @return Value of property firmwareVersion.
     */
    public synchronized String getFirmwareVersion() throws IOException {
        os.write("f".getBytes());
        os.flush();
        String version;
        try {
            watchdog.interruptAfter(TIMEOUT);
            version = r.readLine();
            watchdog.cancelInterrupt();
        } catch (Exception e) {
            throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
        }
        //Add decimal point. (Assume only two places after decimal.)
        version = version.substring(0, version.length() - 2) + "."
                + version.substring(version.length() - 2, version.length());
        return version;
    }
    
    /**
     * Getter for property powerOn.
     * @return Value of property powerOn.
     */
    public synchronized boolean isPowerOn() throws IOException {
        //Determine if power is on.
        //Todo - copy back to main code base
        if (getSensorNames() == null) {
            //Not only is it off, but it hasn't been powered up since
            //being plugged in.
            return false;
        }
        
        double[] lastReading, thisReading;
        lastReading = getInstantGasLevels();
        for (int tries=0; tries < 3; tries++) {
            try {
                Thread.currentThread().sleep(1000);
            } catch (InterruptedException e) {}
            thisReading = getInstantGasLevels();
            for (int i=0; i < thisReading.length; i++) {
                if (lastReading[i] != thisReading[i]) {
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Setter for property powerOn.
     * @param powerOn New value of property powerOn.
     */
    public synchronized void setPowerOn(boolean powerOn) throws IOException {
        boolean isOn = isPowerOn();
        
        if (powerOn == isOn) {
            return;
        }
        
        //Otherwise...
        String response;
        if (powerOn) { //Off to on
            os.write("p".getBytes());
            os.flush();
            
            try {
                watchdog.interruptAfter(TIMEOUT);
                response = r.readLine();
                watchdog.cancelInterrupt();
            } catch (Exception e) {
                throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
            }
            if (response.indexOf("On") < 0) {
                throw new IOException(
                        "Tried to turn MultiRAE Plus on, but got response:\n\t"
                        + response);
            }
            try {
                Thread.currentThread().sleep(97000); //97 seconds to turn on //TODO
            } catch (InterruptedException e) {}
        } else { //On to off
            os.write("p".getBytes());
            os.flush();
            try {
                watchdog.interruptAfter(TIMEOUT);
                response = r.readLine();
                watchdog.cancelInterrupt();
            } catch (Exception e) {
                throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
            }
            if (response.indexOf("Off") < 0) {
                throw new IOException(
                        "Tried to turn MultiRAE Plus off, but got response:\n\t"
                        + response);
            }
            try {
                Thread.currentThread().sleep(1000); //1 second to turn off
            } catch (InterruptedException e) {}
        }
        
//        propertyChangeSupport.firePropertyChange("powerOn",
//                new Boolean(isOn), new Boolean(powerOn));
    }
    
    /**
     * Getter for property serialNumber.
     * @return Value of property serialNumber.
     */
    public synchronized String getSerialNumber() throws IOException {
        os.write("s".getBytes());
        os.flush();
        String sn;
        try {
            watchdog.interruptAfter(TIMEOUT);
            sn = r.readLine();
            watchdog.cancelInterrupt();
        } catch (Exception e) {
            throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
        }
        return sn;
    }
    
    /**
     * Getter for property modelNumber.
     * @return Value of property modelNumber.
     */
    public synchronized String getModelNumber() throws IOException {
        os.write("m".getBytes());
        os.flush();
        String m;
        try {
            watchdog.interruptAfter(TIMEOUT);
            m = r.readLine();
            watchdog.cancelInterrupt();
        } catch (Exception e) {
            throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
        }
        return m;
    }
    
    public void stop() {
        try {
            is.close();
        } catch (Throwable t) {
            t.printStackTrace();
        }
        try {
            os.close();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        watchdog.stop();
        stopPolling();
    }
    
    /**
     * Getter for property sensorNames.
     * @return Value of property sensorNames.
     */
    public String[] getSensorNames() {
        return sensorNames; //which was read in the constructor
    }
    
    /**
     * Loads sensor names from the unit.
     * @return Value of property sensorNames.
     */
    private String[] getSensorNamesFromUnit() throws IOException {
        os.write("n".getBytes());
        os.flush();
        String n;
        try {
            watchdog.interruptAfter(TIMEOUT);
            n = r.readLine();
            watchdog.cancelInterrupt();
        } catch (Exception e) {
            throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
        }
        if (n.indexOf("***") >= 0) { //Todo - copy back to main codebase
            return null; //Rae hasn't yet been powered on, and doesn't know
        }
        return StringSplit.split(n, " ");
    }
    
    /**
     * Getter for property instantGasLevels.
     * @return Value of property instantGasLevels.
     */
    public double[] getInstantGasLevels() throws IOException {
        os.write("r".getBytes());
        os.flush();
        String n;
        try {
            watchdog.interruptAfter(TIMEOUT);
            n = r.readLine();
            watchdog.cancelInterrupt();
        } catch (Exception e) {
            throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
        }
        String[] levels = StringSplit.split(n, " ");
        double[] result = new double[levels.length];
        for (int i=0; i < result.length; i++) {
            result[i] = Double.parseDouble(levels[i]);
            result[i] /= 10;
        }
        return result;
    }
    
    public String toString() {
        String result = null;
        try {
            result = "RAE Systems MultiRAE:\n" +
                    " Model: " + getModelNumber() + "\n" +
                    " Serial number: " + getSerialNumber() + "\n" +
                    " Firmware version: " + getFirmwareVersion() + "\n" +
                    " Sensors: ";
        } catch (IOException e) {
            return e.toString();
        }
        for (int i=0; i < sensorNames.length; i++) {
            result += (sensorNames[i] + " ");
        }
        result += "\n";
        return result;
        
        //Todo - sensor errors
    }
    
    /**
     * Getter for property calculatedReadings.
     * @return Value of property calculatedReadings.
     */
    public CalculatedReading[] getCalculatedReadings() throws IOException {
        CalculatedReading[] readings = new CalculatedReading[sensorNames.length];
        os.write("c".getBytes());
        os.flush();
        String n;
        for (int i=0; i < readings.length; i++) {
            try {
                watchdog.interruptAfter(TIMEOUT);
                n = r.readLine();
                watchdog.cancelInterrupt();
            } catch (Exception e) {
                throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
            }
            String[] reading = StringSplit.split(n, " ");
            if (reading.length == 4) {
                readings[i] = new CalculatedReading(
                        sensorNames[i],
                        Double.parseDouble(reading[0])/10,
                        Double.parseDouble(reading[1])/10,
                        Double.parseDouble(reading[2])/10,
                        Double.parseDouble(reading[3])/10);
            } else {
                readings[i] = new CalculatedReading(
                        sensorNames[i],
                        Double.parseDouble(reading[0])/10,
                        Double.parseDouble(reading[1])/10);
            }
        }
        return readings;
    }
    
    /**
     * Getter for property alarmLimits.
     * @return Value of property alarmLimits.
     */
    public CalculatedReading[] getAlarmLimits() throws IOException {
        CalculatedReading[] limits = new CalculatedReading[sensorNames.length];
        os.write("l".getBytes());
        os.flush();
        String n;
        for (int i=0; i < limits.length; i++) {
            try {
                watchdog.interruptAfter(TIMEOUT);
                n = r.readLine();
                watchdog.cancelInterrupt();
            } catch (Exception e) {
                throw new IOException("Failed to respond in " + TIMEOUT + " ms.");
            }
            String[] limit = StringSplit.split(n, " ");
            if (limit.length == 4) {
                limits[i] = new CalculatedReading(
                        sensorNames[i],
                        Double.parseDouble(limit[0])/10,
                        Double.parseDouble(limit[1])/10,
                        Double.parseDouble(limit[2])/10,
                        Double.parseDouble(limit[3])/10);
            } else {
                limits[i] = new CalculatedReading(
                        sensorNames[i],
                        Double.parseDouble(limit[0])/10,
                        Double.parseDouble(limit[1])/10);
            }
        }
        return limits;
    }
    
    /**
     * Utility field holding list of InstantGasLevelsListeners.
     */
    private transient java.util.Vector InstantGasLevelsListenerList;
    
    /**
     * Registers InstantGasLevelsListener to receive events.
     * @param listener The listener to register.
     */
    public synchronized void addInstantGasLevelsListener(InstantGasLevelsListener listener) {
        
        if (InstantGasLevelsListenerList == null ) {
            InstantGasLevelsListenerList = new java.util.Vector();
        }
        InstantGasLevelsListenerList.addElement(listener);
    }
    
    /**
     * Removes InstantGasLevelsListener from the list of listeners.
     * @param listener The listener to remove.
     */
    public synchronized void removeInstantGasLevelsListener(InstantGasLevelsListener listener) {
        
        if (InstantGasLevelsListenerList != null ) {
            InstantGasLevelsListenerList.removeElement(listener);
        }
    }
    
    /**
     * Notifies all registered listeners about the event.
     *
     * @param event The event to be fired
     */
    private synchronized void fireInstantGasLevelsListenerNewInstantGasLevelsEvent(double[] event) {
        for (int i = 0; i < InstantGasLevelsListenerList.size(); i++) {
            ((InstantGasLevelsListener)InstantGasLevelsListenerList.elementAt(i)).newInstantGasLevels(event);
        }
    }
    
    protected void fireNewEvent() {
        try {
            fireInstantGasLevelsListenerNewInstantGasLevelsEvent(getInstantGasLevels());
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }
    
}